/* vobcopy 0.4.0
 *
 * Copyright (c) 2001, 2002 robos@users.sourceforge.net
 * Lots of contribution and cleanup from rosenauer@users.sourceforge.net
 *
 *
 *  This file is part of vobcopy.
 *
 *  vobcopy is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  vobcopy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with vobcopy; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

/* CONTRIBUTORS (direct or through source "borrowing") 
 * rosenauer@users.sf.net - helps me a lot! 
 * Billy Biggs <vektor@dumbterm.net> - took his play_title.c code and 
 *  implemeted it here 
 * Hkan Hjort <d95hjort@dtek.chalmers.se> and Billy Biggs - libdvdread
 */


   
/* For more information, take a look into dvd_reader.h */

/*CHANGES:
 * -behind -1, -2, -3 and -4 you can specify auxiliary/alternative directories 
 * where the data will be written to. The data will be split to fit on the 
 * first (-o) and split to fit on the second (-1) ...
 * (beware: the path probably has to be specified *directly* behind the number 
 * like this -1/tmp/ and might even then be necessary to to this to -o too)
 * -I gives you information about the titles, chapters and angles on the dvd.
 * -v -v for super-verbosity: makes a logfile called vobcopy-(VERSION).log 
 * (rudimentary only till now) for sending to me: robos@users.sourceforge.net
 * -l for large file output on systems that support it. Testing for supported 
 * systems not yet included
 * - efforts to get vobcopy to work on solaris (it already copies something!)
 * - no options need to be supplied to vobcopy any longer, it's selfseeking 
 * like a sidewinder ;-), but in case you need to specify the device the -i 
 * (like input) option is available
 * - rosenauer@users.sf.net joined me in the effort to make vobcopy virtually
 * perfect ;-)
 * -o /path/to/output-dir/ option
 * -n number_of_title-vob
 * - get dvd title
 * many thanx to nils for his cdtitle.c code (he has to thank someone for that
 too)
 * - progression bar (kinda)
 * - name of outputted files formatted like: name-of-movie1-1.vob, 
 * name-of-movie1-2.vob according to the chosen title and sub-vobs
 * - check size of the combined vobs on dvd against what has been copied (did 
 it copy everything)
 * - check in advance if on the drive the stuff has to be copied to is enough 
 space free
 * - check the provided command line options whether they are what we need and
 not wrong type (not 100% but hey...)
 */

/*TODO:
 * - a detection method if the users system supports large files (> 2GB) or not
 has to be implemented.
 * - error handling with the errno and strerror function
 */

/*THOUGHT-ABOUT FEATURES
 * - being able to specify
     - which title
     - which chapter
     - which angle 
     - which language(s)
     - which subtitle(s)
     to copy
 * - print the total playing time of that title
 * - rip all titles
 * - image the whole dvd
 */

/*If you find a bug, contact me at robos@users.sourceforge.net*/

/* LFS-support is really needed at the moment */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/vfs.h>
#include <dvdread/dvd_reader.h>
/* for solaris, but is also present in linux */
#include <sys/statvfs.h> 
#include <errno.h>
#include "dvd.h"
#include "vobcopy.h"

/*for play_title.c*/
#include <assert.h>
/* #include "config.h" */
#include <dvdread/ifo_types.h>
#include <dvdread/ifo_read.h>
#include <dvdread/dvd_udf.h>
#include <dvdread/nav_read.h>
#include <dvdread/nav_print.h>

extern int errno;

/* from play_title */
/**
 * Returns true if the pack is a NAV pack.  This check is clearly insufficient,
 * and sometimes we incorrectly think that valid other packs are NAV packs.  I
 * need to make this stronger.
 */
int is_nav_pack( unsigned char *buffer )
{
    return ( buffer[ 41 ] == 0xbf && buffer[ 1027 ] == 0xbf );
}

/* --------------------------------------------------------------------------*/
/* MAIN */
/* --------------------------------------------------------------------------*/

int main ( int argc, char *argv[] )
{
  int               streamout;
  char              name[35] = "", op, temp[3], temp1[3];
  char              path_cd[255], logfile_name[20],logfile_path[255];
  char              title_name[33], vobcopy_call[255];
  char              *pointer;
  char              pwd[255],provided_output_dir[255],provided_input_dir[255];
  /*these things might have to be initialised to /dev/null so that they won't 
    write into the wrong space */
  char              alternate_output_dir1[255], alternate_output_dir2[255];
  char              alternate_output_dir3[255], alternate_output_dir4[255];
  unsigned char     bufferin[DVD_VIDEO_LB_LEN];
  int               i = 0,j = 0, l = 0, argc_i = 0, alternate_dir_count = 0;
  int               partcount = 0, title_return, options_char = 0;
  int               cd_count = 0, verbosity_level = 0, paths_taken = 0;
  off_t             pwd_free, vob_size = 0, disk_vob_size = 0;
  off_t             offset = 0, free_space = 0; 
  off_t             max_filesize_in_blocks = 1048571;  /* for 2^31 / 2048 */
  off_t             max_filesize_in_blocks_summed = 0;
  ssize_t           file_size_in_blocks = 0;
  bool              mounted = FALSE, provided_output_dir_flag = FALSE;
  bool              verbose_flag = FALSE, provided_input_dir_flag = FALSE;
  bool              force_flag = FALSE, info_flag = FALSE,still_copying = TRUE;
  bool              large_file_flag = FALSE;
  
  struct stat       buf;
#ifdef linux
  struct statfs     buf1;
#else 
  struct statvfs    buf1; 
#endif 
  dvd_reader_t      *dvd = NULL;
  dvd_file_t        *dvd_file;
  dvd_read_domain_t domain = DVD_READ_TITLE_VOBS ;
  extern char       *optarg;
  extern int        optind, optopt;

  /*this is taken from play_title.c*/

  int titleid = 1, chapid = 0, pgc_id, len, start_cell, cur_cell;
  unsigned int cur_pack;
  int angle = 0, ttn, pgn, next_cell, sum_chapters = 0, sum_angles = 0;
  int most_chapters = 0;
  unsigned char data[ 1024 * DVD_VIDEO_LB_LEN ];
  ifo_handle_t *vmg_file;
  tt_srpt_t *tt_srpt;
  ifo_handle_t *vts_file;
  vts_ptt_srpt_t *vts_ptt_srpt;
  pgc_t *cur_pgc;


  /*
   * See if the correct number if arguments are given
   */
#ifdef __USE_FILE_OFFSET64
  if( argc > 9)
#else
    if( argc > 8)
#endif
      {
	usage(argv[0]);
      }
  
  /*
   * the getopt part (getting the options from command line)
   */
  while (1)
    {
      options_char = getopt(argc, argv, "1:2:3:4:a:c:i:o:n:vflhL:VI");
      if (options_char == -1) break;
      
      switch(options_char)
	{
	  case'a':
	    if (!isdigit(*optarg))
	      { 
		fprintf(stderr, "The thing behind -a has to be a number! \n");
		exit(1); 
	      } 
	  sscanf( optarg, "%i", &angle );
	  angle--;/*in the ifo they start at zero */
	  break;

	  case'c':
	    if (!isdigit(*optarg))
	      { 
		fprintf(stderr, "The thing behind -n has to be a number! \n");
		exit(1); 
	      } 
	  sscanf( optarg, "%i", &chapid );
	  chapid--;/*in the ifo they start at zero */
	  break;

	  case'f':
	    force_flag = TRUE;
	  break;

	  case'h':
	    usage(argv[0]);
	  break;

	  case'i':
	    if (isdigit(*optarg))
	      { 
		fprintf(stderr, "Erm, the number comes behind -n ... \n");
		exit(1); 
	      } 
	  strncpy(provided_input_dir, optarg, 255);
	  provided_input_dir_flag = TRUE;
	  break;

#ifdef __USE_FILE_OFFSET64	    
	  case'l':
	    max_filesize_in_blocks = 4500000000000000; 
	  /* 2^63 / 2048 (not exactly) */
	  large_file_flag = TRUE;
	  break;
#endif

	  case'n':
	    if (!isdigit(*optarg))
	      { 
		fprintf(stderr, "The thing behind -n has to be a number! \n");
		exit(1); 
	      } 
	  sscanf( optarg, "%i", &titleid );
	  break;

	  case'o':
	    if (isdigit(*optarg))
	      { 
		fprintf(stderr, "Erm, the number comes behind -n ... \n");
		exit(1); 
	      } 
	  strncpy(provided_output_dir, optarg, 255);
	  provided_output_dir_flag = TRUE;
	  alternate_dir_count++;
	  force_flag = TRUE;
	  break;
	 
	  case'1':
	    if(alternate_dir_count < 1)
	      { 
		fprintf(stderr, "Please specify output dirs in this order: -o -1 -2 -3 -4 \n");
		exit(1); 
	      } 

	    if (isdigit(*optarg))
	      { 
		fprintf(stderr, "Erm, the number comes behind -n ... \n");
		exit(1); 
	      } 
	  strncpy(alternate_output_dir1, optarg, 255);
	  provided_output_dir_flag = TRUE;
	  alternate_dir_count++;
	  force_flag = TRUE;
	  break;

	  case'2':
	    if(alternate_dir_count < 2)
	      { 
		fprintf(stderr, "Please specify output dirs in this order: -o -1 -2 -3 -4 \n");
		exit(1); 
	      } 

	    if (isdigit(*optarg))
	      { 
		fprintf(stderr, "Erm, the number comes behind -n ... \n");
		exit(1); 
	      } 
	  strncpy(alternate_output_dir2, optarg, 255);
	  provided_output_dir_flag = TRUE;
	  alternate_dir_count++;
	  force_flag = TRUE;
	  break;

	  case'3':
	    if(alternate_dir_count < 3)
	      { 
		fprintf(stderr, "Please specify output dirs in this order: -o -1 -2 -3 -4 \n");
		exit(1); 
	      } 

	    if (isdigit(*optarg))
	      { 
		fprintf(stderr, "Erm, the number comes behind -n ... \n");
		exit(1); 
	      } 
	  strncpy(alternate_output_dir3, optarg, 255);
	  provided_output_dir_flag = TRUE;
	  alternate_dir_count++;
	  force_flag = TRUE;
	  break;

	  case'4':
	    if(alternate_dir_count < 4)
	      { 
		fprintf(stderr, "Please specify output dirs in this order: -o -1 -2 -3 -4 \n");
		exit(1); 
	      } 

	    if (isdigit(*optarg))
	      { 
		fprintf(stderr, "Erm, the number comes behind -n ... \n");
		exit(1); 
	      } 
	  strncpy(alternate_output_dir4, optarg, 255);
	  provided_output_dir_flag = TRUE;
	  alternate_dir_count++;
	  force_flag = TRUE;
	  break;

	  case'v':
	    strcpy(logfile_path, "/tmp/"); 
	  verbose_flag = TRUE;
	  verbosity_level++;
	  break;
	    
	  case'I':
	    info_flag = TRUE;
	  break;
	    
	  case'L':
	    strncpy(logfile_path, optarg, 255);	       
	  strcat(logfile_path, "/");
	  verbose_flag = TRUE;
	  verbosity_level = 2;
	  break;
	    
	  case'V':
	    printf("Vobcopy "VERSION" - GPL Copyright (c) 2001, 2002 robos@users.sourceforge.net\n");
	  exit(0);
	  break;
	    
	  case':':
	    if (optopt == 49)
	      {
	        printf("-1 is missing its path argument \n");
	        exit(1);
	      }

	    if (optopt == 50)
	      {
	        printf("-2 is missing its path argument \n");
	        exit(1);
	      }

	    if (optopt == 51)
	      {
	        printf("-3 is missing its path argument \n");
	        exit(1);
	      }

	    if (optopt == 52)
	      {
	        printf("-4 is missing its path argument \n");
	        exit(1);
	      }


	    if (optopt == 97)
	      {
	        printf("-a is missing its path argument \n");
	        exit(1);
	      }

	  if (optopt == 99)
	    {
	      printf("-c is missing its path argument \n");
	      exit(1);
	    }

	  if (optopt == 105)
	    {
	      printf("-i is missing its path argument \n");
	      exit(1);
	    }
	    
	  if (optopt == 110)
	    {
	      printf("-n is missing the title-number \n");
	      exit(1);
	    }
	    
	  if (optopt == 111)
	    {
	      printf("-o is missing its path argument \n");
	      exit(1);
	    }

	  if (optopt == 76)
	    {
	      printf("-L is missing its path argument \n");
	      exit(1);
	    }
	    
	  break;
	    
	  case'?':
	    fprintf(stderr, "Wrong option.\n");
	  usage(argv[0]);
	  exit(1);
	  break;
	    
	default:
	  fprintf(stderr, "Wrong option.\n");
	  usage(argv[0]);
	  exit(1);
	}
    }

  if(verbosity_level > 1)
    {
      printf("Uhu, super-verbose\n");
      strcpy(logfile_name, logfile_path);
      strcat(logfile_name, "vobcopy_");
      strcat(logfile_name, VERSION);
      strcat(logfile_name, ".log");
      printf("The log-file is written to %s\n", logfile_name);
      printf("If you don't like that position, use -L /path/to/logfile/ instead of -v -v\n");
      stderr = fopen(logfile_name, "a");
      strcpy(vobcopy_call, argv[0]);
      for(argc_i = 1; argc_i != argc; argc_i++)
	{
	  strcat(vobcopy_call, argv[argc_i]);
	  strcat(vobcopy_call, " ");
	}
      fprintf(stderr, "--------------------------------------------------------------------------------");
      fprintf(stderr, "Called: %s\n", vobcopy_call);
    }
  
  /*
   * Check if the provided path is too long
   */
  if ( optind < argc ) /* there is still the path as in vobcopy-0.2.0 */
    {
      provided_input_dir_flag = TRUE;
      if (strlen( argv[optind] ) >= 255)
	{
	  fprintf(stderr, "\nBloody path to long '%s'\n", argv[optind]);
	  exit(1);
	}
      strncpy( provided_input_dir, argv[optind],255 );
    }

  if (provided_input_dir_flag) /*the path has been given to us */
    {   
      if ((mounted = get_device( provided_input_dir, path_cd )) < 0)
	{
	  fprintf(stderr, "Could not get the device which belongs to the given path!\n");
	  exit(1);
	}
    }
  else /*need to get the path and device ourselves (oyo = on your own) */
    {
      if ((cd_count = get_device_oyo( provided_input_dir, path_cd )) <= 0)
	{
	  fprintf(stderr, "Could not get the device and path! Maybe not mounted the dvd?\n");
	  exit(1);
	}
      if(cd_count > 0)
	mounted = TRUE;
    }

  /* 
   * Is the path correct
   */
  printf("\npath to cd: %s\n", path_cd);
  if(verbosity_level > 1)
    fprintf(stderr, "\npath to cd: %s\n", path_cd);
   
  if(!( dvd = DVDOpen( path_cd ))) 
    {
      fprintf(stderr, "\nPath thingy didn't work '%s'\n", argv[optind]);
      fprintf(stderr, "Try someting like -i /dev/cdrom, /dev/dvd, /dvd or even /cdrom\n");
      if(cd_count > 1) 
	fprintf(stderr, "By the way, you have %i cdroms/dvds mounted, that probably caused the problem\n", cd_count);
      DVDClose( dvd );
      exit(1);
    }
   
  title_return = get_cdromtitle(path_cd, title_name);

  /* The new part taken from play-title.c*/

  /**
   * Load the video manager to find out the information about the titles on
   * this disc.
   */
  vmg_file = ifoOpen( dvd, 0 );
  if( !vmg_file ) {
    fprintf( stderr, "Can't open VMG info.\n" );
    DVDClose( dvd );
    return -1;
  }
  tt_srpt = vmg_file->tt_srpt;

  /**
   * Make sure our title number is valid.
   */
  fprintf( stderr, "There are %d titles on this DVD.\n",
	   tt_srpt->nr_of_srpts );
  if( titleid <= 0 || (titleid-1) >= tt_srpt->nr_of_srpts ) {
    fprintf( stderr, "Invalid title %d.\n", titleid );
    ifoClose( vmg_file );
    DVDClose( dvd );
    return -1;
  }


  /**
   * Make sure the chapter number is valid for this title.
   */
  for(i = 0;i < tt_srpt->nr_of_srpts;i++)
    {
      sum_chapters += tt_srpt->title[ i ].nr_of_ptts;
      if(i > 0)
	{
	  if(tt_srpt->title[ i ].nr_of_ptts > tt_srpt->title[ most_chapters ].nr_of_ptts)
	    {
	      most_chapters = i;
	    }
	}
      
    }

  fprintf( stderr, "There are %i chapters on the dvd.\n", sum_chapters );
  fprintf( stderr, "Most chapters has title %i with %d chapters.\n", 
	   (most_chapters + 1), tt_srpt->title[ most_chapters ].nr_of_ptts );

  if( chapid < 0 || chapid >= tt_srpt->title[ titleid-1 ].nr_of_ptts ) {
    fprintf( stderr, "Invalid chapter %d\n", chapid + 1 );
    ifoClose( vmg_file );
    DVDClose( dvd );
    return -1;
  }


  /**
   * Make sure the angle number is valid for this title.
   */
  for(i = 0;i < tt_srpt->nr_of_srpts;i++)
    {
      sum_angles += tt_srpt->title[ i ].nr_of_angles;
    }

  fprintf( stderr, "There are %d angles on this dvd.\n", sum_angles );
  if( angle < 0 || angle >= tt_srpt->title[ titleid-1 ].nr_of_angles ) {
    fprintf( stderr, "Invalid angle %d\n", angle + 1 );
    ifoClose( vmg_file );
    DVDClose( dvd );
    return -1;
  }

   
  /*
   * get the whole vob size via stat() manually
   */
  if(mounted) 
    {
      vob_size = get_vobsize(titleid, provided_input_dir);
    } 
  
  /* 
   * check if on the target directory is enough space free
   * see man statfs for more info
   */
  
  /*get the current working directory*/
  if (provided_output_dir_flag)
    {
      strcpy( pwd, provided_output_dir);
    }
  else
    {
      if ( getcwd( pwd, 255) == NULL )
	{ 
	  fprintf(stderr, "\nHmm, the path length of your current directory is really large (>255)\n");
	  fprintf(stderr, "Change to a path with shorter path length pleeeease ;-)\n");
	  exit(1);   
	}
    }
  
  /* add a trailing '/' to pwd */
/*   if (pwd[strlen(pwd)-1] != '/') */
/*     { */
/*       pointer = pwd + strlen(pwd); */
/*       *pointer = '/'; */
/*       pointer++; */
/*       *pointer = '\0'; */
/*     } */
  end_slash_adder(pwd);
  
  
  /* get available space on target filesystem */
/* #ifdef linux */
/*   statfs(pwd, &buf1); */
/*   if(verbosity_level >= 1) */
/*     fprintf(stderr,"Used the linux statfs\n"); */
/* #else */
/*   statvfs(pwd, &buf1); */
/*   if(verbosity_level >= 1) */
/*     fprintf(stderr,"Used statvfs\n"); */
/* #endif */
/*   pwd_free = buf1.f_bavail * buf1.f_bsize; */
  pwd_free = freespace_getter(pwd,verbosity_level);
  
  /*
   * Open now up the actual files for reading
   * they come from libdvdread merged together under the given title number
   * (thx again for the great library)
   */

  printf("Using Title: %i\n", titleid);
  printf("Title has %d chapters and %d angles\n",tt_srpt->title[ titleid - 1 ].nr_of_ptts,tt_srpt->title[ titleid - 1 ].nr_of_angles );

  printf("Using Chapter: %i\n", chapid + 1);
  printf("Using Angle: %i\n", angle + 1);
  if(verbosity_level > 1)
    {
      fprintf(stderr,"Using Title: %i\n", titleid);
      fprintf(stderr,"Title has %d chapters and %d angles\n",tt_srpt->title[ titleid - 1 ].nr_of_ptts,tt_srpt->title[ titleid - 1 ].nr_of_angles );
      fprintf(stderr,"Using Chapter: %i\n", chapid + 1);
      fprintf(stderr,"Using Angle: %i\n", angle + 1);
    }

  
  if(info_flag && vob_size != 0)
    {
      fprintf(stderr,"\nDVD-title: %s\n", title_name);
      fprintf(stderr, "  disk free: %lld MB\n", pwd_free / (1024*1024) );
      fprintf(stderr, "  vobs size: %lld MB\n", vob_size / (1024*1024) ); 
      ifoClose( vmg_file );
      DVDCloseFile( dvd_file );
      DVDClose( dvd );
      /*hope all are closed now...*/
      exit(0);
    }
  

  /**
   * Load the VTS information for the title set our title is in.
   */
  vts_file = ifoOpen( dvd, tt_srpt->title[ titleid-1 ].title_set_nr );
  if( !vts_file ) {
    fprintf( stderr, "Can't open the title %d info file.\n",
	     tt_srpt->title[ titleid-1 ].title_set_nr );
    ifoClose( vmg_file );
    DVDClose( dvd );
    return -1;
  }


  /**
   * Determine which program chain we want to watch.  This is based on the
   * chapter number.
   */
  ttn = tt_srpt->title[ titleid-1 ].vts_ttn;
  vts_ptt_srpt = vts_file->vts_ptt_srpt;
  pgc_id = vts_ptt_srpt->title[ ttn - 1 ].ptt[ chapid ].pgcn;
  pgn = vts_ptt_srpt->title[ ttn - 1 ].ptt[ chapid ].pgn;
  cur_pgc = vts_file->vts_pgcit->pgci_srp[ pgc_id - 1 ].pgc;
  start_cell = cur_pgc->program_map[ pgn - 1 ] - 1;


  /*   if ((dvd_file = DVDOpenFile( dvd, titleid, domain)) == NULL) */
  /*     { */
  /*       fprintf(stderr, "\nDVDOpenFile failed\n"); */
  /*       fprintf(stderr, "Wrong title number?\n"); */
  /*       DVDClose( dvd ); */
  /*       exit(1); */
  /*     } */


  /**
   * We've got enough info, time to open the title set data.
   */
  dvd_file = DVDOpenFile( dvd, tt_srpt->title[ titleid-1 ].title_set_nr,
			  DVD_READ_TITLE_VOBS );
  if( !dvd_file ) {
    fprintf( stderr, "Can't open title VOBS (VTS_%02d_1.VOB).\n",
	     tt_srpt->title[ titleid-1 ].title_set_nr );
    ifoClose( vts_file );
    ifoClose( vmg_file );
    DVDClose( dvd );
    return -1;
  }



  file_size_in_blocks = DVDFileSize( dvd_file );

  
  if (vob_size == 0)
    {
      vob_size = (off_t) (file_size_in_blocks) * (off_t) DVD_VIDEO_LB_LEN;
      if(verbosity_level >= 1)
	fprintf(stderr, "vob_size was 0\n");
    }
  

  /*debug-output: difference between vobsize read from cd and size returned from libdvdread */
  if (mounted && verbose_flag)
    {
      printf("\ndifference between vobsize read from cd and size returned from libdvdread:\n");
      printf("vob_size (stat) = %lld\nlibdvdsize = %lld\ndiff = %lld\n",  
	     vob_size, 
	     (off_t) (file_size_in_blocks) * (off_t) DVD_VIDEO_LB_LEN, 
	     vob_size - (off_t) (file_size_in_blocks) * (off_t) DVD_VIDEO_LB_LEN );
      if(verbosity_level > 1)
	{
	  fprintf(stderr,"\ndifference between vobsize read from cd and size returned from libdvdread:\n");
	  fprintf(stderr,"vob_size (stat) = %lld\nlibdvdsize = %lld\ndiff = %lld\n", 
		  vob_size, 
		  (off_t) (file_size_in_blocks) * (off_t) DVD_VIDEO_LB_LEN, 
		  vob_size - (off_t) (file_size_in_blocks) * (off_t) DVD_VIDEO_LB_LEN );
	}
    }

  if(info_flag)
    {
      fprintf(stderr,"\nDVD-title: %s\n", title_name);
      fprintf(stderr, "  disk free: %lld MB\n", pwd_free / (1024*1024) );
      fprintf(stderr, "  vobs size: %lld MB\n", vob_size / (1024*1024) ); 
      ifoClose( vmg_file );
      DVDCloseFile( dvd_file );
      DVDClose( dvd );
      /*hope all are closed now...*/
      exit(0);
    }

  
  /* now the actual check if enough space is free*/
  if ( pwd_free < vob_size)
    {
      fprintf(stderr, "\n  disk free: %lld MB", pwd_free / (1024*1024) );
      fprintf(stderr, "\n  vobs size: %lld MB", vob_size / (1024*1024) ); 
      fprintf(stderr, "\nHmm, better change to a dir with enough space left or call with -f (force) \n");
      if(pwd_free == 0 && !force_flag)
	{
	  if(verbosity_level > 1)
	    fprintf(stderr,"Hmm, statfs (statvfs) seems not to work on that directory. \n");
	  printf("Hmm, statfs (statvfs) seems not to work on that directory. \n");
	  printf("Nevertheless, do you want vobcopy to continue [y] or do you want to check for \n");
	  printf("enough space first [q]?\n");
	  
	  while (1)
	    {
	      op=fgetc( stdin );
	      if(op == 'y')
		{
		  force_flag = TRUE;
		  if(verbosity_level >= 1)
		    fprintf(stderr, "y pressed - force write\n");
		  break;
		}
	      else if(op == 'n' || op =='q')
		{
		  if(verbosity_level >= 1)
		    fprintf(stderr, "n/q pressed\n");
		  exit(1);
		  break;
		}
	      else
		{
		  printf("please choose [y] to continue or [n] to quit\n");
		}
	    }
	}
      if (!force_flag)	  
	exit(1);
    }
  


/* Handschuhe                                                                                                                                                    Es war einmal ein junger Mann, der in die Stadt ging, um ein                   Geburtstagsgeschenk f?r seine neue Freundin zu erwerben. Da die beiden noch    nicht sehr lange zusammen waren, beschloss er - nach reiflicher ?berlegung -   ihr ein paar Handschuhe zu kaufen. Ein romantisches, aber doch nicht zu        pers?nliches Geschenk.                                                         In Begleitung der j?ngeren Schwester seiner Freundin ging er zu C&A und        erstand ein paar weisse Handschuhe. Die Schwester kaufte ein Unterh?schen      f?r sich. Beim Einpacken vertauschte die Verk?uferin aus Versehen die          Sachen. So bekam die Schwester die Handschuhe eingepackt und der junge Mann    bekam unwissend das Paket mit dem H?schen, das er auf dem R?ckweg zur Post     brachte und mit einem kleinen Briefchen an seine Liebste verschickte:                                                                                         "Mein Schatz, ich habe mich f?r dieses Geschenk entschieden, da ich            festgestellt habe, dass Du keine tr?gst, wenn wir abends zusammen ausgehen.    Wenn es nach mir gegangen w?re, h?tte ich mich f?r die langen mit den          Kn?pfen entschieden, aber deine Schwester meinte, die kurzen w?ren besser.     Sie tr?gt sie auch und man kriegt sie leichter aus. Ich weiss, dass das eine   empfindliche Farbe ist, aber die Dame, bei der ich sie gekauft habe, zeigte    mir ihre, die sie nun seit drei Wochen tr?gt, und sie waren ?berhaupt nicht    schmutzig. Ich bat sie, Deine f?r mich anzuprobieren und sie sah echt Klasse   darin aus. Ich w?nschte, ich k?nnte sie Dir beim ersten Mal anziehen, aber     ich denke, bis wir uns wiedersehen werden sie mit einer Menge anderer H?nde    in Ber?hrung gekommen sein. Wenn Du sie ausziehst, vergiss nicht, kurz         hinein zu blasen bevor Du sie weglegst, da sie wahrscheinlich ein bisschen     feucht vom Tragen werden. Denk immer daran, wie oft ich sie in Deinem          kommenden Lebensjahr k?ssen werde. Ich hoffe, Du wirst sie Freitagabend f?r    mich tragen.                                                                                                                                                  In Liebe                                                                                                                                                      PS: Der letzte Schrei ist, sie etwas hochgekrempelt zu tragen, so dass der     Pelz rausguckt!" */




  /********************* 
   * this is the main read and copy loop
   *********************/
  printf("\nDVD-title: %s\n", title_name);
  if(verbosity_level > 1)
    fprintf(stderr,"\nDVD-title: %s\n", title_name);
  while(offset < file_size_in_blocks)
    {
/*       strcpy(name, pwd); */
/*       if(!title_return) */
/*         { */
/* 	  strcat( name, title_name); */
/*         } */
/*       else */
/* 	strcat( name, "the_video_you_wanted"); */
      
/*       sprintf(temp, "%d", titleid); */
/*       strcat(name, temp); */
/*       strcat(name, "-"); */
      partcount++; 
/*       sprintf(temp1, "%d", partcount); */
/*       strcat(name, temp1); */
/*       strcat(name, ".vob"); */
/*       printf("\nOutputting to %s\n", name); */
/*       if(verbosity_level > 1) */
/* 	fprintf(stderr,"Outputting to %s\n", name); */

      /*new part to distribute the files over different directories*/
      if(paths_taken == 0)
	{
	  end_slash_adder(provided_output_dir);
	  free_space = freespace_getter(provided_output_dir,verbosity_level);
	  if(verbosity_level > 1)
	    fprintf(stderr,"free space for -o dir: %d\n",free_space);
	  output_path_maker(pwd,name,title_return,title_name,titleid, partcount);
	}
      else
	{
	  /*j--;*/ /* needs to adjust the count for the main copy loop */
	  if(alternate_dir_count == 1) /*shouldn't come to this else only if 
					some data is still there but no free 
					space left... */
	    {
	      fprintf(stderr, "Hmm, looks like there wasn't enough space in the provided dirs...\n");
	      fprintf(stderr, "Did you use -1, -2, -3 or even up to -4 to specify alternate dirs?\n");
	    }

	  if(paths_taken == 1 && alternate_dir_count > 1)
	    {
	      end_slash_adder(alternate_output_dir1);
	      free_space = freespace_getter(alternate_output_dir1,verbosity_level);
	      if(verbosity_level > 1)
		fprintf(stderr,"free space for -1 dir: %lld\n",free_space);

	      output_path_maker(alternate_output_dir1,name,title_return,
				title_name,titleid,partcount);
	      alternate_dir_count--;
	    }

	  if(paths_taken == 2 && alternate_dir_count > 1)
	    {
	      end_slash_adder(alternate_output_dir2);
	      free_space = freespace_getter(alternate_output_dir2,verbosity_level);
	      if(verbosity_level > 1)
		fprintf(stderr,"free space for -2 dir: %lld\n",free_space);
	      output_path_maker(alternate_output_dir2,name,title_return,title_name,titleid,partcount);
	      alternate_dir_count--;
	    }

	  if(paths_taken == 3 && alternate_dir_count > 1)
	    {
	      end_slash_adder(alternate_output_dir3);
	      free_space = freespace_getter(alternate_output_dir3,verbosity_level);
	      if(verbosity_level > 1)
		fprintf(stderr,"free space for -3 dir: %lld\n",free_space);
	      output_path_maker(alternate_output_dir3,name,title_return,title_name,titleid,partcount);
	      alternate_dir_count--;
	    }

	  if(paths_taken == 4 && alternate_dir_count > 1)
	    {
	      end_slash_adder(alternate_output_dir4);
	      free_space = freespace_getter(alternate_output_dir4,verbosity_level);
	      if(verbosity_level > 1)
		fprintf(stderr,"free space for -4 dir: %lld\n",free_space);
	      output_path_maker(alternate_output_dir4,name,title_return,title_name,titleid,partcount);
	    }
	}
      /*here the output size gets adjusted to the given free space*/

      if(!large_file_flag && force_flag && free_space < 2147473408) /* 2GB */
	{
	  max_filesize_in_blocks = ((free_space - 2097152) / 2048); /* - 2 MB */
	  if(verbosity_level > 1)
	    fprintf(stderr,"taken max_filesize_in_blocks(2GB version): %lld\n",max_filesize_in_blocks);
	  paths_taken++;
	}
      if(large_file_flag && force_flag) /*lfs version */
	{
	  max_filesize_in_blocks = ((free_space - 2097152) / 2048);/* - 2 MB */
	  if(verbosity_level > 1)
	    fprintf(stderr,"taken max_filesize_in_blocks(lfs version): %lld\n",max_filesize_in_blocks);
	  paths_taken++;
	}
      
#ifdef __USE_LARGEFILE64
      if( open(name, O_RDONLY|O_LARGEFILE) >= 0)
#else
      if( open(name, O_RDONLY) >= 0)
#endif
	{
	  printf("\nFile '%s' already exists, [o]verwrite, [a]ppend, [q]uit? \n", name);
          if(verbosity_level > 1)
	    fprintf(stderr,"\nFile '%s' already exists, [o]verwrite, [a]ppend, [q]uit? \n", name);
	  while (1)
	    {
	      op=fgetc( stdin );
	      fgetc (stdin ); /* probably need to do this for second time it 
				 comes around this loop */
	      if(op == 'o')
		{
#ifdef __USE_LARGEFILE64
		  if( (streamout = open(name, O_WRONLY | O_TRUNC | O_LARGEFILE)) < 0)
#else
		  if( (streamout = open(name, O_WRONLY | O_TRUNC )) < 0)
#endif
		  {
		      fprintf(stderr, "\nerror open file %s\n", name);
		      exit (1);
		  }
		  break;
		}
	      else if(op == 'a')
		{
#ifdef __USE_LARGEFILE64
		  if( (streamout = open(name, O_WRONLY | O_APPEND | O_LARGEFILE)) < 0)
#else
		  if( (streamout = open(name, O_WRONLY | O_APPEND)) < 0)
#endif
		  {
		      fprintf(stderr, "\nerror open file %s\n", name);
		      exit (1);
		  }
		  if(verbosity_level >= 1)
		    fprintf(stderr, "user chose append\n");
		  break;
		}
	      else if(op == 'q')
		{
		  DVDCloseFile(dvd_file);
		  DVDClose( dvd );
		  exit(1);
		}
	      else
		{
		  printf("\nplease choose [o]verwrite, [a]ppend, [q]uit the next time ;-)\n");
		}
	    }
	}
      else
	{
	  /*assign the stream */
#ifdef __USE_LARGEFILE64
	  if( (streamout = open(name, O_WRONLY | O_CREAT | O_LARGEFILE, 0644)) < 0)
#else
	  if( (streamout = open(name, O_WRONLY | O_CREAT, 0644)) < 0)
#endif
	  {
	      fprintf(stderr, "\nerror open file %s\n", name);
	      exit (1);
	  }
	}
      
      /* this here is the main copy part */	
      
      printf("\n");
      memset( bufferin, 0, BLOCK_COUNT * DVD_VIDEO_LB_LEN * sizeof(unsigned char));

/*       for (; offset < (off_t) file_size_in_blocks && offset - (off_t) j * max_filesize_in_blocks < max_filesize_in_blocks; offset++) */
      for (; offset < (off_t) file_size_in_blocks && offset - (off_t)max_filesize_in_blocks_summed < max_filesize_in_blocks; offset++)
	{
	  DVDReadBlocks( dvd_file, offset, BLOCK_COUNT, bufferin);
	  if(write( streamout, bufferin, DVD_VIDEO_LB_LEN) < 0)
	  {
	      fprintf(stderr, "\nwrite() error\n"
		              "it's possible that you try to write files\n"
			      "greater than 2GB to filesystem which\n"
			      "doesn't support it? (try without -l)\n");
	      fprintf(stderr, "error: %s\n",strerror(errno));
	      exit(1);
	  }
	  l++;
	  if( l == 100 )
	    {
	      /*this is the progress indicator*/
	      printf("%8lld of%8lld Blocks written (%lld %%)\r", 
		    offset, 
		    (off_t) file_size_in_blocks,
	            (offset*100)/file_size_in_blocks );
	      l=0;
	    }
	}

      close(streamout);  
      if(verbosity_level >= 1)
	{
	  fprintf(stderr,"max_filesize_in_blocks %8lld \n",max_filesize_in_blocks);
	  fprintf(stderr,"offset at the end %8lld \n", offset);
	  fprintf(stderr,"file_size_in_blocks %8lld \n",(off_t) file_size_in_blocks );
	}
      /* now lets see whats the size of this file in bytes */
      stat(name, &buf);
      disk_vob_size += (off_t) buf.st_size;
      if(verbosity_level >= 1)
	{
	  fprintf(stderr, "Single file size (of copied file) %lld\n", buf.st_size);
	  fprintf(stderr, "Cumulated size %lld\n", disk_vob_size);
	}
      
      max_filesize_in_blocks_summed += max_filesize_in_blocks;
      /* j++; */	/* # of seperate files we have written */
    }
  if(verbosity_level >= 1)
    fprintf(stderr, "# of separate files: %i\n", j);
  
  /*
   * clean up and close everything 
   */

  ifoClose( vts_file );
  ifoClose( vmg_file );
  DVDCloseFile( dvd_file );
  DVDClose( dvd );
  printf("\nCopying finished! Let's see if the sizes match (roughly)\n");   
  printf("Combined size of title-vobs: %lld (%lld MB)\n", vob_size, vob_size / (1024*1024));
  printf("Copied size (size on disk):  %lld (%lld MB)\n", disk_vob_size, disk_vob_size / (1024*1024));
  if(verbosity_level > 1)
    {
      fprintf(stderr,"\nCopying finished! Let's see if the sizes match (roughly)\n");   
      fprintf(stderr,"Combined size of title-vobs: %lld (%lld MB)\n", vob_size, vob_size / (1024*1024));
      fprintf(stderr,"Copied size (size on disk):  %lld (%lld MB)\n", disk_vob_size, disk_vob_size / (1024*1024));
    }
  
  if (( vob_size - disk_vob_size ) > MAX_DIFFER )
    {
      printf("Hmm, the sizes differ by more than %d\n", MAX_DIFFER);
      printf("Take a look with MPlayer if the output is ok\n");
      if(verbosity_level > 1)
	fprintf(stderr,"Hmm, the sizes differ by more than %d\n", MAX_DIFFER);
    }
  else
    {
      printf("Everything seems to be fine, the sizes match pretty good ;-)\n");
      printf("Have a lot of fun!\n");
    }
  
  return 0;
}

void usage(char *program_name){
  fprintf(stderr, "\nUsage: %s [-i /path/to/the/mounted/dvd/] [-n title-number] \n",program_name);
  fprintf(stderr, "[-o /path/to/output-dir/] [-f (force output)] [-V (version)] [-v (verbose)]\n");
  fprintf(stderr, "[-v -v (create log-file)] [-h (this here ;-)] \n");
  fprintf(stderr, "[-I (infos about title, chapters and angles on the dvd)]\n");
  fprintf(stderr, "[-1/path/to/second/output/dir/] [-2/.../third/..] [-3/../] [-4 /../]\n");
#ifdef __USE_FILE_OFFSET64
  fprintf(stderr, "[-l (large-file support for files > 2GB) (still somewhat experimental)] \n");
#endif
  exit(1);
}


int end_slash_adder(char *path)  
{  /* add a trailing '/' to path */
  char *pointer;
  if (path[strlen(path)-1] != '/')
    {
      pointer = path + strlen(path);
      *pointer = '/';
      pointer++;
      *pointer = '\0';
    }
  return 0;
}


off_t freespace_getter(char *path, int verbosity_level)
{
    /* get available space on target filesystem */
#ifdef linux
  struct statfs     buf1;
#else 
  struct statvfs    buf1; 
#endif 
/*   ssize_t temp1, temp2; */
  long temp1, temp2;
  off_t sum;
#ifdef linux
  statfs(path, &buf1);
  if(verbosity_level >= 1)
    fprintf(stderr,"Used the linux statfs\n");
#else
  statvfs(path, &buf1);
  if(verbosity_level >= 1)
    fprintf(stderr,"Used statvfs\n");
#endif
  temp1 = buf1.f_bavail;
  temp2 = buf1.f_bsize;
  sum = ((off_t)temp1 * (off_t)temp2);
  if(verbosity_level >= 1)
    {
      fprintf(stderr,"In freespace_getter:for %s : %lld free\n", path, sum);
      fprintf(stderr,"In freespace_getter:part1 %ld, part2 %ld\n", temp1, temp2);
    }
  /*   return (buf1.f_bavail * buf1.f_bsize); */
  return sum;
}

int output_path_maker(char *pwd,char *name,int title_return, char *title_name,int titleid, int partcount)
{
  char temp[3];
  strcpy(name, pwd);
  if(!title_return)
    {
      strcat( name, title_name);
    }
  else
    strcat( name, "the_video_you_wanted");
  
  sprintf(temp, "%d", titleid);
  strcat(name, temp);
  strcat(name, "-");
  sprintf(temp, "%d", partcount);
  strcat(name, temp);
  strcat(name, ".vob");
  printf("\nOutputting to %s\n", name);
  return 0;
}









